/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.common;

import java.net.URI;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.JBIException;
import javax.jbi.component.Component;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ComponentLifeCycle;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.ObjectName;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.objectweb.petals.component.common.listener.MessageExchangeListener;
import org.objectweb.petals.component.common.listener.MessageExchangeProcessor;
import org.objectweb.petals.component.common.listener.ProcessingResult;
import org.objectweb.petals.component.common.listener.ThreadPoolConfiguration;
import org.objectweb.petals.component.common.su.ServiceUnitExtensionsValidator;
import org.objectweb.petals.component.common.su.ServiceUnitListener;
import org.objectweb.petals.component.common.su.SimpleServiceUnitManager;
import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.component.common.util.PetalsExtensionsUtil;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.objectweb.petals.common.descriptor.Consumes;
import org.objectweb.petals.common.descriptor.Extensions;
import org.objectweb.petals.common.descriptor.Provides;
import org.objectweb.petals.common.util.StringHelper;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;

/**
 * An abstract class that can be used to easily create a new JBI Component. It
 * implements Component, ComponentLifeCycle and MessageExchangeProcessor JBI
 * interfaces.
 *
 * @author ofabre
 *
 */
public abstract class AbstractComponent implements Component,
        ComponentLifeCycle, MessageExchangeProcessor {

    private static final String MESSAGE_EXCHANGE_SETTING_ERROR = "Error during message exchange setting";

    /**
     * The component's delivery channel
     */
    private DeliveryChannel channel;

    /**
     * The Component's Context
     */
    private ComponentContext context;

    /**
     * The extension properties. The property file has been created during the
     * bootstrap phase
     */
    protected ComponentPropertiesManager extensionProperties;

    /**
     * The message exchange listener
     */
    private MessageExchangeListener listener;

    /**
     * The component's logger
     */
    private Logger log; // NOPMD by ofabre

    /**
     * The ServiceUnit lifecycle listener
     */
    private ServiceUnitListener serviceUnitListener;

    /**
     * The service unit extensions validator
     */
    private ServiceUnitExtensionsValidator serviceUnitExtensionsValidator;

    /**
     * The component's service unit manager
     */
    private SimpleServiceUnitManager suManager;

    /**
     * The configuration of thread pool used by the component to process
     * messages.
     */
    private ThreadPoolConfiguration tpConfiguration;

    /**
     * InOnly Mep shortcut
     */
    public final static String IN_ONLY = "in-only";

    /**
     * InOnly Mep shortcut
     */
    public final static String ROBUST_IN_ONLY = "robust-in-only";

    /**
     * InOnly Mep shortcut
     */
    public final static String IN_OUT = "in-out";

    /**
     * InOnly Mep shortcut
     */
    public final static String IN_OPTIONAL_OUT = "in-opt-out";

    /**
     * If the received exchange is a fault response from the consumer,
     * acknowledge it : status DONE
     *
     * @param exchange
     *            The processed {@link MessageExchange}
     * @return true if the ack. of the Fault is done; false if no fault is
     *         present in the exchange
     * @throws MessagingException
     */
    private boolean ackFaultReception(final MessageExchangeWrapper exchange)
            throws MessagingException {
        boolean result = false;

        if (exchange.getFault() != null) {
            exchange.setDoneStatus();
            result = true;
        }
        return result;
    }

    /**
     * Creates a Fault from an exception
     *
     * @param e
     *            The {@link Exception} used to create a new {@link Fault}
     * @param exchange
     *            The {@link MessageExchange} that allows to create a new
     *            {@link Fault}
     * @return the created Fault
     * @throws PEtALSComponentSDKException
     *             if an error occured during Fault creation or filling
     */
    private Fault createFault(final JBIException e,
            final MessageExchangeWrapper exchange)
            throws PEtALSComponentSDKException {
        String msg = "";
        Fault f = null;
        try {
            f = exchange.createFault();
        } catch (MessagingException e1) {
            msg = "Error during Fault creation";
            log.severe(msg);
            throw new PEtALSComponentSDKException(msg, e1);
        }
        // String faultString = SourceHelper.createSoapFault(e, null);
        String faultString = SourceHelper.createSoapFault(e, "0");
        Source content = SourceHelper.createSource(faultString);
        try {
            f.setContent(content);
        } catch (MessagingException e1) {
            msg = "Error during Fault content setting";
            log.severe(msg);
            throw new PEtALSComponentSDKException(msg, e1);
        }
        return f;
    }

    /**
     * Creates a new MessageExchangeListener
     *
     * @param processor
     *            the messageExchange processor
     * @param tpConfig
     *            The configuration of the thread pool used by the listener to
     *            process messages
     * @return the created {@link MessageExchangeListener}
     */
    private MessageExchangeListener createMessageExchangeListener(
            final MessageExchangeProcessor processor,
            final ThreadPoolConfiguration tpConfig) {
        return new MessageExchangeListener(channel, processor, tpConfig,
                MessageExchangeListener.IgnoredStatus.DONE_AND_ERROR_IGNORED,
                0, context.getComponentName(), log);
    }

    /**
     * Can be overidden to provide a specific service unit manager
     *
     * @return the service unit manager used by this component.
     */
    protected SimpleServiceUnitManager createServiceUnitManager() {
        return new SimpleServiceUnitManager(this, context, log);
    }

    /**
     * Can be overidden to provide a specific way to configure the thread pool.
     *
     * @return the thread pool configuration.
     */
    protected ThreadPoolConfiguration createThreadPoolConfiguration() {
        ThreadPoolConfiguration conf = new ThreadPoolConfiguration();
        conf.setCorePoolSize(Integer.parseInt(extensionProperties
                .getAttributeValue(ComponentPropertiesManager.CORE_POOL_SIZE)));

        conf
                .setKeepAliveTime(Long
                        .parseLong(extensionProperties
                                .getAttributeValue(ComponentPropertiesManager.KEEP_ALIVE_TIME)));

        conf.setMaxPoolSize(Integer.parseInt(extensionProperties
                .getAttributeValue(ComponentPropertiesManager.MAX_POOL_SIZE)));

        conf.setQueueSize(Integer.parseInt(extensionProperties
                .getAttributeValue(ComponentPropertiesManager.QUEUE_SIZE)));

        this.log.info("Thread pool configuration -> Queue size : "
                + conf.getQueueSize() + ", Core pool size : "
                + conf.getCorePoolSize() + ", Max pool size : "
                + conf.getMaxPoolSize() + ", Keep alive : "
                + conf.getKeepAliveTime() + " ms");

        return conf;
    }

    /**
     * Returns the component delivery channel, used to send messages to petals
     * container
     *
     * @return the delivery channel. Not null.
     */
    public DeliveryChannel getChannel() {
        return channel;
    }

    /**
     * The component context
     *
     * @return the component context. Not null.
     */
    public ComponentContext getContext() {
        return context;
    }

    private Extensions getExchangeEpJBIExtensions(
            final MessageExchangeWrapper exchange) {
        ServiceEndpoint ep = exchange.getEndpoint();
        Extensions extensions = null;
        if (ep != null) {
            Provides provides = suManager.getProvidesFromEndpoint(ep);
            if (provides != null) {
                extensions = provides.getExtensions();
            }
        }
        return extensions;
    }

    /**
     * Returns null by default as this class does not provides JMX extension
     * MBean
     *
     * @see javax.jbi.component.ComponentLifeCycle#getExtensionMBeanName()
     */
    public ObjectName getExtensionMBeanName() {
        return null;
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.Component#getLifeCycle()
     */
    public ComponentLifeCycle getLifeCycle() {
        return this;
    }

    /**
     * Returns the component logger
     *
     * @return component logger. Not null.
     */
    public Logger getLogger() {
        return log;
    }

    /**
     * Can be overidden to return a custom resource bundle name
     *
     * @return the resource bundle name or null if no resource bundle name has
     *         to be used (default value)
     */
    protected String getResourceBundleName() {
        return null;
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.Component#getServiceDescription(javax.jbi.servicedesc.ServiceEndpoint)
     */
    public Document getServiceDescription(ServiceEndpoint endpoint) {
        return suManager.getServiceDescription(endpoint);
    }

    public ServiceUnitListener getServiceUnitListener() {
        return serviceUnitListener;
    }

    public ServiceUnitExtensionsValidator getServiceUnitExtensionsValidator() {
        return serviceUnitExtensionsValidator;
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.Component#getServiceUnitManager()
     */
    public ServiceUnitManager getServiceUnitManager() {
        return suManager;
    }

    /**
     * This method is called at the end of the component classical
     * initialization. It could be overidden to make additional initialization.
     * When this method is called, the component's delivery channel, context,
     * logger and service unit manager are available.
     *
     */
    protected void init() throws JBIException {

    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.ComponentLifeCycle#init(javax.jbi.component.ComponentContext)
     */
    public void init(ComponentContext compContext) throws JBIException {
        this.context = compContext;
        this.log = context.getLogger("", getResourceBundleName());
        channel = context.getDeliveryChannel();
        suManager = createServiceUnitManager();
        extensionProperties = loadProperties();
        tpConfiguration = createThreadPoolConfiguration();
        log.log(Level.INFO, "Component initialized");
        init();
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.Component#isExchangeWithConsumerOkay(javax.jbi.servicedesc.ServiceEndpoint,
     *      javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithConsumerOkay(ServiceEndpoint endpoint,
            MessageExchange exchange) {
        return true;
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.Component#isExchangeWithProviderOkay(javax.jbi.servicedesc.ServiceEndpoint,
     *      javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithProviderOkay(ServiceEndpoint endpoint,
            MessageExchange exchange) {
        return true;
    }

    /**
     * Load the extensions properties
     */
    protected ComponentPropertiesManager loadProperties() {
        ComponentPropertiesManager extensionProperties = new ComponentPropertiesManager();
        extensionProperties.setRootPath(getContext().getInstallRoot());
        extensionProperties.load();
        return extensionProperties;
    }

    /**
     * Called to process a message exchange. MessageExchange handles all
     * information of the exchange between provider and consumer (operation,
     * MessageExchangePattern, payload...). Optionnal extensions can be used to
     * handle complementary information
     *
     * @param exchange
     *            the message exchange to process, wrapped to add some util
     *            methods. Not null
     * @param extensions
     *            JBI descriptor extensions. The message exchange targets a
     *            specific endpoint, this endpoint has been activated during the
     *            processing of a provides node of a service unit jbi
     *            descriptor. Theses extensions are the extensions of this
     *            provides node. Can be null
     * @return true if the out response has been set, false otherwise (Useful
     *         for a in-out or in optionnal out mep)
     *
     * @see MEPConstants
     * @throws HandlingException
     *             a Fault will be generated with the exception, the "out"
     *             message will not be treated
     */
    protected abstract boolean onExchange(MessageExchangeWrapper exchange,
            Extensions extensions) throws HandlingException;

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.listener.MessageExchangeProcessor#process(org.objectweb.petals.component.common.util.MessageExchangeWrapper)
     */
    public ProcessingResult process(final MessageExchangeWrapper exchange)
            throws JBIException {
        if (exchange == null) {
            throw new JBIException("Message exchange to process can't be null");
        }
        if (exchange.isActiveStatus()) {
            if (exchange.isProviderRole()) {
                if (exchange.isInOnlyPattern()) {
                    processInOnly(exchange);
                } else if (exchange.isRobustInOnlyPattern()) {
                    processRobustInOnly(exchange);
                } else if (exchange.isInOutPattern()
                        || exchange.isInOptionalOutPattern()) {
                    processInOut(exchange);
                } else {
                    MessagingException e = new MessagingException(
                            "MessageExchangePattern not recognized :"
                                    + exchange.getPattern());
                    exchange.setError(e);
                    channel.send(exchange.getMessageExchange());
                    throw e;
                }
                channel.send(exchange.getMessageExchange());
            }
        } else {
            log.log(Level.INFO, "No processing on Unactive MEP");
        }
        return ProcessingResult.SUCESS;
    }

    /**
     * Process "in-only" MessageExchange
     *
     * @param exchange
     *            the MessageExchange
     * @throws PEtALSComponentSDKException
     *             When errors occur during processing
     */
    private void processInOnly(final MessageExchangeWrapper exchange)
            throws PEtALSComponentSDKException {
        try {
            Extensions extensions = getExchangeEpJBIExtensions(exchange);
            onExchange(exchange, extensions);
        } catch (HandlingException e) {
            throw new PEtALSComponentSDKException(
                    "Processing error but InOnly message can not contains Fault.",
                    e);
        } finally {
            try {
                if (exchange.isActiveStatus()) {
                    exchange.setDoneStatus();
                }
            } catch (MessagingException e) {
                log.severe(MESSAGE_EXCHANGE_SETTING_ERROR);
                throw new PEtALSComponentSDKException(
                        MESSAGE_EXCHANGE_SETTING_ERROR, e);
            }
        }
    }

    /**
     * Process "in-out" MessageExchange
     *
     * @param exchange
     *            the MessageExchange
     * @throws PEtALSComponentSDKException
     *             When errors occur during processing
     *
     */
    private void processInOut(final MessageExchangeWrapper exchange)
            throws PEtALSComponentSDKException {
        try {
            if (!ackFaultReception(exchange)) {

                Extensions extensions = getExchangeEpJBIExtensions(exchange);

                boolean outTreated = onExchange(exchange, extensions);

                if (!outTreated) {
                    exchange.setDoneStatus();
                }
            }
        } catch (HandlingException e) {
            try {
                log.log(Level.SEVERE,
                        "Error during message exchange processing", e);
                exchange.setFault(createFault(e, exchange));
            } catch (MessagingException e1) {
                log.severe(MESSAGE_EXCHANGE_SETTING_ERROR);
                throw new PEtALSComponentSDKException(
                        MESSAGE_EXCHANGE_SETTING_ERROR, e1);
            }
        } catch (MessagingException e) {
            log.severe(MESSAGE_EXCHANGE_SETTING_ERROR);
            throw new PEtALSComponentSDKException(
                    MESSAGE_EXCHANGE_SETTING_ERROR, e);
        }
    }

    /**
     * Process "robust-in-only" MessageExchange
     *
     * @param exchange
     *            the MessageExchange
     * @throws PEtALSComponentSDKException
     *             When errors occur during processing
     *
     */
    private void processRobustInOnly(final MessageExchangeWrapper exchange)
            throws PEtALSComponentSDKException {
        try {
            Extensions extensions = getExchangeEpJBIExtensions(exchange);
            onExchange(exchange, extensions);
        } catch (HandlingException e) {
            try {
                log.log(Level.SEVERE,
                        "Error during message exchange processing", e);
                exchange.setFault(createFault(e, exchange));
            } catch (MessagingException e1) {
                log.severe(MESSAGE_EXCHANGE_SETTING_ERROR);
                throw new PEtALSComponentSDKException(
                        MESSAGE_EXCHANGE_SETTING_ERROR, e1);
            }
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.Component#resolveEndpointReference(org.w3c.dom.DocumentFragment)
     */
    public ServiceEndpoint resolveEndpointReference(DocumentFragment epr) {
        return null;
    }

    /**
     * Send a message to the endpoint associated to the mapped address. For
     * exemple, this method can be used by a listener, that listen to messages
     * from external services (linked to the given address), in order to send
     * these messages to internal JBI Components.
     *
     * @param address
     *            the distant address of the external consumer service. Must be
     *            non null and non empty.
     * @param exchange
     *            the message exchange to be sent to the internal JBI Component.
     *            Must be non null.
     * @return true if the message has been sent (if a timeout occurs, return
     *         false)
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message on the container
     */
    public boolean sendMessage(final String address,
            final MessageExchange exchange) throws MessagingException {
        boolean sent = false;

        if (StringHelper.isNullOrEmpty(address)) {
            throw new MessagingException(
                    "You must provide a non null and non empty address");
        }
        if (exchange == null) {
            throw new MessagingException(
                    "You must provide a non null MessageExchange");
        }

        // retrieve timeout
        Consumes consumes = suManager.getConsumesFromAddress(address);
        Extensions extensions = consumes.getExtensions();
        long timeout = Long.parseLong(PetalsExtensionsUtil
                .extractValueFromKeyValueExtension(extensions,
                        ExtensionsConstants.TIMEOUT, "0"));

        ServiceEndpoint ep = null;
        if (consumes.getEndpointName() != null
                && consumes.getServiceName() != null) {
            ep = context.getEndpoint(consumes.getServiceName(), consumes
                    .getEndpointName());
        }

        exchange.setInterfaceName(consumes.getInterfaceName());
        exchange.setService(consumes.getServiceName());
        exchange.setEndpoint(ep);

        try {
            sent = channel.sendSync(exchange, timeout);
        } catch (JBIException e) {
            throw new MessagingException(
                    "Failed to send the mapped message to the delivery channel.",
                    e);
        }
        return sent;
    }

    /**
     * Send a message to the endpoint associated to the mapped address. The
     * {@link MessageExchange} is created from information given in parameters
     * (content, attachments, properties, operation and pattern). If no
     * operation and pattern are specified, it tries to find it in the
     * "consumes" extensions associated to the given address. If no pattern can
     * be found, InOnly pattern is used by default.
     *
     * @param address
     *            the distant address of the external consumer service. Must be
     *            non null and non empty.
     * @param content
     *            the {@link Source} content of the "IN" message. Can be null
     * @param attachments
     *            a {@link Map} of "IN" message attachments. Can be null or
     *            empty
     * @param properties
     *            a {@link Map} of "IN" message properties. Can be null or empty
     * @param operation
     *            the targeted operation. Can be null
     * @param pattern
     *            the {@link MessageExchange} pattern. Can be null
     * @return the {@link MessageExchange} that has been sent
     * @throws MessagingException
     */
    public MessageExchange sendMessage(final String address,
            final Source content, final Map<String, DataHandler> attachments,
            Map<String, Object> properties, QName operation, URI pattern)
            throws MessagingException {
        boolean sent = false;

        if (StringHelper.isNullOrEmpty(address)) {
            throw new MessagingException(
                    "You must provide a non null and non empty address");
        }

        MessageExchange exchange = null;

        Consumes consumes = suManager.getConsumesFromAddress(address);
        if (consumes != null) {
            Extensions extensions = consumes.getExtensions();
            exchange = createMessageExchange(MEPConstants.valueOf(pattern),
                    extensions);
            if (operation == null) {
                operation = QName.valueOf(PetalsExtensionsUtil
                        .extractValueFromKeyValueExtension(extensions,
                                ExtensionsConstants.OPERATION));
            }
            exchange.setOperation(operation);
        } else {
            throw new MessagingException(
                    "Can't retrieve consumes service unit "
                            + "element matching the given address");
        }

        setInMessage(exchange, content, attachments, properties);

        try {
            sent = sendMessage(address, exchange);
        } catch (JBIException e) {
            throw new MessagingException(e);
        }

        if (!sent) {
            throw new MessagingException(
                    "Failed to send the mapped message to the delivery channel,"
                            + " sendSync() method returned false");
        }

        return exchange;
    }

    /**
     * Send a message to the endpoint associated to the mapped address. The
     * {@link MessageExchange} is created from information given in parameters
     * (content, attachments, properties).
     *
     * @param address
     *            the distant address of the external consumer service. Must be
     *            non null and non empty.
     * @param content
     *            the {@link Source} content of the "IN" message. Can be null
     * @param attachments
     *            a {@link Map} of "IN" message attachments. Can be null or
     *            empty
     * @param properties
     *            a {@link Map} of "IN" message properties. Can be null or empty
     * @return the {@link MessageExchange} that has been sent
     * @throws MessagingException
     */
    public MessageExchange sendMessage(final String address,
            final Source content, final Map<String, DataHandler> attachments,
            Map<String, Object> properties) throws MessagingException {
        return sendMessage(address, content, attachments, properties, null,
                null);
    }

    /**
     * Send a message to the endpoint associated to the mapped address. The
     * {@link MessageExchange} is created from information given in parameters
     * (content, attachments).
     *
     * @param address
     *            the distant address of the external consumer service. Must be
     *            non null and non empty.
     * @param content
     *            the {@link Source} content of the "IN" message. Can be null
     * @param attachments
     *            a {@link Map} of "IN" message attachments. Can be null or
     *            empty
     * @return the {@link MessageExchange} that has been sent
     * @throws MessagingException
     */
    public MessageExchange sendMessage(final String address,
            final Source content, final Map<String, DataHandler> attachments)
            throws MessagingException {
        return sendMessage(address, content, attachments, null, null, null);
    }

    /**
     * Send a message to the endpoint associated to the mapped address. The
     * {@link MessageExchange} is created from information given in parameters
     * (content).
     *
     * @param address
     *            the distant address of the external consumer service. Must be
     *            non null and non empty.
     * @param content
     *            the {@link Source} content of the "IN" message. Can be null
     * @return the {@link MessageExchange} that has been sent
     * @throws MessagingException
     */
    public MessageExchange sendMessage(final String address,
            final Source content) throws MessagingException {
        return sendMessage(address, content, null, null, null, null);
    }

    private void setInMessage(MessageExchange exchange, Source content,
            Map<String, DataHandler> attachments, Map<String, Object> properties)
            throws MessagingException {
        NormalizedMessage message = exchange.createMessage();
        message.setContent(content);
        if (attachments != null) {
            for (String attachment : attachments.keySet()) {
                message.addAttachment(attachment, attachments.get(attachment));
            }
        }
        if (properties != null) {
            for (String property : properties.keySet()) {
                message.setProperty(property, properties.get(property));
            }
        }
        exchange.setMessage(message, "in");
    }

    /**
     * Create a message exchange matching the given pattern. If pattern
     * parameter is null, try to find a specified pattern in "consumes"
     * extensions. Finaly, if no pattern can be found, an InOnly exchange is
     * created by default.
     *
     * @param mep
     *            a specific MEP pattern
     * @param extensions
     *            "consumes" {@link Extensions}
     * @return a message exchange matching the given pattern
     * @throws MessagingException
     *             if the message exchange can't be created or if the specified
     *             pattern isn't valid
     */
    private MessageExchange createMessageExchange(MEPConstants mep,
            Extensions extensions) throws MessagingException {
        if (mep == null) {
            mep = retrievePattern(extensions);
        }
        MessageExchange exchange = null;
        if (mep.equals(MEPConstants.IN_ONLY_PATTERN)) {
            exchange = channel.createExchangeFactory().createInOnlyExchange();
        } else if (mep.equals(MEPConstants.IN_OPTIONAL_OUT_PATTERN)) {
            exchange = channel.createExchangeFactory()
                    .createInOptionalOutExchange();
        } else if (mep.equals(MEPConstants.IN_OUT_PATTERN)) {
            exchange = channel.createExchangeFactory().createInOutExchange();
        } else if (mep.equals(MEPConstants.ROBUST_IN_ONLY_PATTERN)) {
            exchange = channel.createExchangeFactory()
                    .createRobustInOnlyExchange();
        }
        return exchange;
    }

    /**
     * Retrieve the MEP from the extensions.
     *
     * @param extensions
     * @return
     * @throws MessagingException
     */
    private MEPConstants retrievePattern(Extensions extensions)
            throws MessagingException {
        MEPConstants mep = null;

        String patternShortcut = PetalsExtensionsUtil
                .extractValueFromKeyValueExtension(extensions,
                        ExtensionsConstants.MEP);
        if (!StringHelper.isNullOrEmpty(patternShortcut)) {
            if (ROBUST_IN_ONLY.equalsIgnoreCase(patternShortcut)) {
                mep = MEPConstants.ROBUST_IN_ONLY_PATTERN;
            } else if (IN_OPTIONAL_OUT.equalsIgnoreCase(patternShortcut)) {
                mep = MEPConstants.IN_OPTIONAL_OUT_PATTERN;
            } else if (IN_OUT.equalsIgnoreCase(patternShortcut)) {
                mep = MEPConstants.IN_OUT_PATTERN;
            } else if (IN_ONLY.equalsIgnoreCase(patternShortcut)) {
                mep = MEPConstants.IN_ONLY_PATTERN;
            } else {
                throw new MessagingException("MEP pattern specified in "
                        + "consumes extension (" + patternShortcut
                        + ") isn't valid. " + "Valid patterns are : in-only, "
                        + "in-opt-out, robust-in-only, in-out.");
            }
        } else {
            mep = MEPConstants.IN_ONLY_PATTERN;
        }
        return mep;
    }

    public void setServiceUnitListener(ServiceUnitListener serviceUnitListener) {
        this.serviceUnitListener = serviceUnitListener;
    }

    public void setServiceUnitExtensionsValidator(
            ServiceUnitExtensionsValidator serviceUnitExtensionsValidator) {
        this.serviceUnitExtensionsValidator = serviceUnitExtensionsValidator;
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.ComponentLifeCycle#shutDown()
     */
    public void shutDown() throws JBIException {
        log.log(Level.INFO, "shutdown");
        listener = null; // NOPMD frees the memory while component is
        // shutdown
        context = null; // NOPMD idem
        log = null; // NOPMD idem
        channel = null; // NOPMD idem
        suManager = null; // NOPMD idem
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.ComponentLifeCycle#start()
     */
    public void start() throws JBIException {
        log.log(Level.INFO, "start");
        listener = createMessageExchangeListener(this, this.tpConfiguration);
        listener.listen();
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.ComponentLifeCycle#stop()
     */
    public void stop() throws JBIException {
        log.log(Level.INFO, "stop");
        // FIXME : call terminate or close?
        this.listener.terminate();
    }
}
