/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.common.listener;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessagingException;

import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.component.common.util.MessageExchangeWrapperImpl;

/**
 * This class helps a Component to listen MessageExchange on the DeliveryChannel
 * (accept() method). It acts as a Listener. When the listener calls the
 * 'accept()' method, its thread is blocked, but not the object that
 * instantiated it. The listener calls a 'process(Exchange)' on the
 * MessageExchangeProcessor, which acts as a handler of MessageExchanges. This
 * MessageExchangeProcessor has to be provided. <p/
 * <p>
 * The listener can be parametrized with a few parameters. See constructors.
 * </p>
 * 
 * @version $Revision: $ $Date: $
 * @since Petals 1.0
 * @author alouis
 */
public class MessageExchangeListener extends Thread {

    public enum IgnoredStatus {
        NOTHING_IGNORED, DONE_IGNORED, ERROR_IGNORED, DONE_AND_ERROR_IGNORED
    };

    private DeliveryChannel channel;

    private boolean running;

    private IgnoredStatus ignoredStatus;

    private long period;

    private MessageExchangeProcessor processor;

    private Logger logger; // NOPMD by gblondelle

    private LinkedBlockingQueue<Runnable> workers;

    private ThreadPoolExecutor executor;

    /**
     * creates a listener on the specified channel. all exchanges are treated.
     * no interval between two 'accept()';
     * 
     * @param channel
     * @param processor
     * @param tpConfig
     *            The configuration of the thread pool used by this listener to
     *            process messages
     */
    public MessageExchangeListener(DeliveryChannel channel,
        MessageExchangeProcessor processor,
        final ThreadPoolConfiguration tpConfig) {
        this(channel, processor, tpConfig, IgnoredStatus.NOTHING_IGNORED);
    }

    /**
     * creates a listener on the specified channel. exchanges are ignored
     * depending on the specified exchange-status to ignore. no interval between
     * two 'accept()';
     * 
     * @param channel
     * @param processor
     * @param tpConfig
     *            The configuration of the thread pool used by this listener to
     *            process messages
     * @param ignoredStatus
     */
    public MessageExchangeListener(DeliveryChannel channel,
        MessageExchangeProcessor processor,
        final ThreadPoolConfiguration tpConfig, IgnoredStatus ignoredStatus) {
        this(channel, processor, tpConfig, ignoredStatus, 0);
    }

    /**
     * creates a listener on the specified channel. exchanges are ignored
     * depending on the specified exchange-status to ignore. wait for the
     * specified interval between two 'accept()';
     * 
     * @param channel
     * @param processor
     * @param tpConfig
     *            The configuration of the thread pool used by this listener to
     *            process messages
     * @param ignoredStatus
     * @param period
     */
    public MessageExchangeListener(DeliveryChannel channel,
        MessageExchangeProcessor processor,
        final ThreadPoolConfiguration tpConfig, IgnoredStatus ignoredStatus,
        long period) {
        this(channel, processor, tpConfig, ignoredStatus, period, "", null);
    }

    /**
     * creates a listener on the specified channel. exchanges are ignored
     * depending on the specified exchange-status to ignore. wait for the
     * specified interval between two 'accept()';
     * 
     * @param channel
     * @param processor
     * @param tpConfig
     *            The configuration of the thread pool used by this listener to
     *            process messages
     * @param ignoredStatus
     * @param componentName
     *            Name of the component. Used to define the Thread name.
     * @param period
     */
    public MessageExchangeListener(DeliveryChannel channel,
        MessageExchangeProcessor processor,
        final ThreadPoolConfiguration tpConfig, IgnoredStatus ignoredStatus,
        long period, String componentName, Logger log) {
        super(componentName + "-JBI listener thread");
        this.ignoredStatus = ignoredStatus;
        this.channel = channel;
        this.period = period;
        this.processor = processor;
        if (log != null) {
            this.logger = log;
        } else {
            this.logger = LogManager.getLogManager().getLogger(componentName);
        }

        this.workers = new LinkedBlockingQueue<Runnable>(tpConfig
            .getQueueSize());
        this.executor = new ThreadPoolExecutor(tpConfig.getCorePoolSize(),
            tpConfig.getMaxPoolSize(), tpConfig.getKeepAliveTime(),
            TimeUnit.SECONDS, workers);
    }

    /**
     * Close the listener. This cause the terminaison of the loop on
     * DeliveryChannel.accept(), and interrupt this listening thread if the
     * thread is blocked on DeliveryChannel.accept();
     */
    public void close() {
        terminate();
        interrupt();
    }

    /**
     * Start the thread listener ( Thread.start()). The listener waits on
     * DeliveryChannel.accept(), treats the received MessageExchange, and waits
     * again on DeliveryChannel.accept().
     */
    public void listen() {
        start();
    }

    /**
     * loop on accept
     * Stop on Interrupted Exception
     */
    public void run() {
        running = true;

        while (running) {

            // receive a message
            try {
                MessageExchangeWrapper exchange = new MessageExchangeWrapperImpl(
                    channel.accept());

                // forward the message to a worker
                if (!isIgnored(exchange)) {
                    MessageExchangeWorker msgWorker = new MessageExchangeWorker(
                        exchange, this.processor, this.logger);
                    try {
                        executor.execute(msgWorker);
                    } catch (RejectedExecutionException e) {
                        // As the catched RejectedExecutionException has no
                        // message,
                        // we created a new one with a message.
                        // TODO: A resource file should be used to store the
                        // message for localization purpose.
                        RejectedExecutionException ree = new RejectedExecutionException(
                            "No more worker available.", e);
                        logger.log(Level.SEVERE, e.getMessage(), ree);
                        exchange.setError(ree);
                        channel.send(exchange.getMessageExchange());
                    }
                }

                // wait before next accept
                if (period > 0) {
                    try {
                        wait(period);
                    } catch (InterruptedException e) {
                        // nothing
                    }
                }

            } catch (MessagingException e) {
            	// search for an interrupted exception
            	Throwable cause=e;
            	while (cause != null){
            		cause = cause.getCause();
            		if(cause instanceof InterruptedException ){
            			running = false;
            			logger.log(Level.SEVERE, "Thread interrupted, stop listening.");
            		}
            	}
                logger.log(Level.SEVERE, e.getMessage(), e);
            } 
        }
    }

    /**
     * Terminate the listening. On the next MessageExchange treated after the
     * DeliveryChannel.accept(), the listener will stop waiting on
     * DeliveryChannel.accept().
     */
    public void terminate() {
        running = false;

        // Stop workers that are working
        this.executor.shutdownNow();
    }

    /**
     * depending on the IgnoredStatus, tells if the exchange has to be ignored.
     * If the exchange is null, return true.
     * 
     * @param exchange
     * @return
     */
    protected boolean isIgnored(final MessageExchangeWrapper exchange) {
        boolean result = false;

        if (exchange.getMessageExchange() == null) {
            result = true;
        } else {
            if (exchange.isDoneStatus()) {
                result = IgnoredStatus.DONE_IGNORED == ignoredStatus
                    || IgnoredStatus.DONE_AND_ERROR_IGNORED == ignoredStatus;
            } else if (exchange.isErrorStatus()) {
                result = IgnoredStatus.ERROR_IGNORED == ignoredStatus
                    || IgnoredStatus.DONE_AND_ERROR_IGNORED == ignoredStatus;
            }
        }
        return result;
    }
}
